# NOTE: If some tests fail (or take too long)
# you need to tweak the NEATClassifier() and NEATRegressor() init parameters in _neat.py!

from sklearn.utils.estimator_checks import parametrize_with_checks
from neuro_evolution import NEATClassifier
from neuro_evolution import NEATRegressor

@parametrize_with_checks([NEATRegressor(), NEATClassifier()])
def test_sklearn_compatible_estimator(estimator, check):
    check(estimator)


import neat
from neat.config import DefaultClassConfig, ConfigParameter
from neat.math_util import mean
import numpy as np
from ._nsga_sort import nsga_sort
import math
import random

class WannReproduction(neat.DefaultReproduction):
    def __init__(self, config, reporters, stagnation):
        super().__init__(config, reporters,stagnation)

    @classmethod
    def parse_config(cls, param_dict):
        return DefaultClassConfig(param_dict,
                                  [ConfigParameter('elitism', int, 0),
                                   ConfigParameter('survival_threshold', float, 0.2),
                                   ConfigParameter('min_species_size', int, 2),
                                   ConfigParameter('moo_rate', float, 0.80)])


    def reproduce(self, config, species, pop_size, generation):
        """
        Handles creation of genomes, either from scratch or by sexual or
        asexual reproduction from parents.
        """
        # TODO: I don't like this modification of the species and stagnation objects,
        # because it requires internal knowledge of the objects.

        # Filter out stagnated species, collect the  set of non-stagnated
        # species members, and compute their average adjusted fitness.
        # The average adjusted fitness scheme (normalized to the interval
        # [0, 1]) allows the use of negative fitness values without
        # interfering with the shared fitness scheme.
        all_fitnesses = []
        remaining_species = []
        for stag_sid, stag_s, stagnant in self.stagnation.update(species, generation):
            if stagnant:
                self.reporters.species_stagnant(stag_sid, stag_s)
            else:
                all_fitnesses.extend(m.fitness for m in stag_s.members.values())
                remaining_species.append(stag_s)
        # The above comment was not quite what was happening - now getting fitnesses
        # only from members of non-stagnated species.

        # No species left.
        if not remaining_species:
            species.species = {}
            return {} # was []

        # Find minimum/maximum fitness across the entire population, for use in
        # species adjusted fitness computation.
        min_fitness = min(all_fitnesses)
        max_fitness = max(all_fitnesses)
        # Do not allow the fitness range to be zero, as we divide by it below.
        # TODO: The ``1.0`` below is rather arbitrary, and should be configurable.
        fitness_range = max(1.0, max_fitness - min_fitness)
        for afs in remaining_species:
            # Compute adjusted fitness.
            msf = mean([m.fitness for m in afs.members.values()])
            af = (msf - min_fitness) / fitness_range
            afs.adjusted_fitness = af

        adjusted_fitnesses = [s.adjusted_fitness for s in remaining_species]
        avg_adjusted_fitness = mean(adjusted_fitnesses) # type: float
        self.reporters.info("Average adjusted fitness: {:.3f}".format(avg_adjusted_fitness))

        # Compute the number of new members for each species in the new generation.
        previous_sizes = [len(s.members) for s in remaining_species]
        min_species_size = self.reproduction_config.min_species_size
        # Isn't the effective min_species_size going to be max(min_species_size,
        # self.reproduction_config.elitism)? That would probably produce more accurate tracking
        # of population sizes and relative fitnesses... doing. TODO: document.
        min_species_size = max(min_species_size, self.reproduction_config.elitism)
        spawn_amounts = self.compute_spawn(adjusted_fitnesses, previous_sizes,
                                           pop_size, min_species_size)



        new_population = {}
        species.species = {}
        for spawn, s in zip(spawn_amounts, remaining_species):
            # If elitism is enabled, each species always at least gets to retain its elites.
            spawn = max(spawn, self.reproduction_config.elitism)

            assert spawn > 0

            # The species has at least one member for the next generation, so retain it.
            old_members = list(s.members.items())

            meanFit = np.asarray([ind[1].fitness for ind in old_members])
            maxFit = np.asarray([ind[1].fitMax for ind in old_members])
            nConns = np.asarray([len(ind[1].connections) for ind in old_members])
            nConns[ nConns == 0 ] = 1# No connections is pareto optimal but boring...
            objVals = np.c_[meanFit, maxFit, 1/nConns]  # Maximize

            # Rank according to mean performance and max performance
            if self.reproduction_config.moo_rate < np.random.rand():
                rank = nsga_sort(objVals[ :, [0, 1] ])
            else:  # Rank according to mean performance and the number of connections
                rank = nsga_sort(objVals[ :, [0, 2] ])

            # Assign ranks to genomes
            for i in range(len(old_members)):
                old_members[ i ][1].rank = rank[ i ]

            s.members = {}
            species.species[s.key] = s

            # Sort members in order of ascending rank
            old_members.sort(key=lambda x: x[1].rank)

            # Transfer elites to new generation.
            if self.reproduction_config.elitism > 0:
                for i, m in old_members[:self.reproduction_config.elitism]:
                    new_population[i] = m
                    spawn -= 1

            if spawn <= 0:
                continue

            # Only use the survival threshold fraction to use as parents for the next generation.
            repro_cutoff = int(math.ceil(self.reproduction_config.survival_threshold *
                                         len(old_members)))
            # Use at least two parents no matter what the threshold fraction result is.
            repro_cutoff = max(repro_cutoff, 2)
            old_members = old_members[:repro_cutoff]

            # Randomly choose parents and produce the number of offspring allotted to the species.
            while spawn > 0:
                spawn -= 1

                parent1_id, parent1 = random.choice(old_members)
                parent2_id, parent2 = random.choice(old_members)

                # Note that if the parents are not distinct, crossover will produce a
                # genetically identical clone of the parent (but with a different ID).
                gid = next(self.genome_indexer)
                child = config.genome_type(gid)
                child.configure_crossover(parent1, parent2, config.genome_config)
                child.mutate(config.genome_config)
                new_population[gid] = child
                self.ancestors[gid] = (parent1_id, parent2_id)

        return new_population


"""Sklearn NEAT
"""

# Author: Marko Mlakar <markomlakar2@gmail.com>
# License: BSD-3-Clause License

from abc import ABCMeta, abstractmethod
from sklearn.metrics import accuracy_score, r2_score
import neat
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.multiclass import unique_labels
from ._config import NEATConfig, ConfigParser, GenomeConfig, ReproductionConfig, StagnationConfig, SpeciesConfig
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from ._wann_genome import WannGenome
from ._wann_reproduce import WannReproduction


def softmax(x):
    """Compute softmax values for each sets of scores in x"""
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)


class BaseNEAT(BaseEstimator, metaclass=ABCMeta):
    """Base class for NEAT classification and regression.

    Warning: This class should not be used directly. Use derived classes instead!

    """

    @abstractmethod
    def __init__(self, single_shared_weights, number_of_generations, fitness_criterion, fitness_threshold, pop_size,
                 reset_on_extinction, no_fitness_termination, activation_default, activation_mutate_rate,
                 activation_options,
                 aggregation_default, aggregation_mutate_rate, aggregation_options, bias_init_mean,
                 bias_init_stdev, bias_max_value, bias_min_value, bias_mutate_power, bias_mutate_rate,
                 bias_replace_rate, compatibility_disjoint_coefficient, compatibility_weight_coefficient,
                 conn_add_prob, conn_delete_prob, enabled_default, enabled_mutate_rate, feed_forward,
                 initial_connection, node_add_prob, node_delete_prob, num_hidden,
                 response_init_mean, response_init_stdev, response_max_value, response_min_value,
                 response_mutate_power, response_mutate_rate, response_replace_rate, weight_init_mean,
                 weight_init_stdev, weight_max_value, weight_min_value, weight_mutate_power,
                 weight_mutate_rate, weight_replace_rate, compatibility_threshold, species_fitness_func,
                 max_stagnation, species_elitism, elitism, survival_threshold, statistic_reporter,
                 create_checkpoints, checkpoint_frequency, add_conn_roulette_prob, add_node_roulette_prob,
                 mutate_conn_roulette_prob, mutate_act_roulette_prob, delete_conn_roulette_prob,
                 delete_node_roulette_prob):
        self.single_shared_weights = single_shared_weights
        self.number_of_generations = number_of_generations
        self.fitness_criterion = fitness_criterion
        self.fitness_threshold = fitness_threshold
        self.pop_size = pop_size
        self.reset_on_extinction = reset_on_extinction
        self.no_fitness_termination = no_fitness_termination
        self.activation_default = activation_default
        self.activation_mutate_rate = activation_mutate_rate
        self.activation_options = activation_options
        self.aggregation_default = aggregation_default
        self.aggregation_mutate_rate = aggregation_mutate_rate
        self.aggregation_options = aggregation_options
        self.bias_init_mean = bias_init_mean
        self.bias_init_stdev = bias_init_stdev
        self.bias_max_value = bias_max_value
        self.bias_min_value = bias_min_value
        self.bias_mutate_power = bias_mutate_power
        self.bias_mutate_rate = bias_mutate_rate
        self.bias_replace_rate = bias_replace_rate
        self.compatibility_disjoint_coefficient = compatibility_disjoint_coefficient
        self.compatibility_weight_coefficient = compatibility_weight_coefficient
        self.conn_add_prob = conn_add_prob
        self.conn_delete_prob = conn_delete_prob
        self.enabled_default = enabled_default
        self.enabled_mutate_rate = enabled_mutate_rate
        self.feed_forward = feed_forward
        self.initial_connection = initial_connection
        self.node_add_prob = node_add_prob
        self.node_delete_prob = node_delete_prob
        self.num_hidden = num_hidden
        self.response_init_mean = response_init_mean
        self.response_init_stdev = response_init_stdev
        self.response_max_value = response_max_value
        self.response_min_value = response_min_value
        self.response_mutate_power = response_mutate_power
        self.response_mutate_rate = response_mutate_rate
        self.response_replace_rate = response_replace_rate
        self.weight_init_mean = weight_init_mean
        self.weight_init_stdev = weight_init_stdev
        self.weight_max_value = weight_max_value
        self.weight_min_value = weight_min_value
        self.weight_mutate_power = weight_mutate_power
        self.weight_mutate_rate = weight_mutate_rate
        self.weight_replace_rate = weight_replace_rate
        self.compatibility_threshold = compatibility_threshold
        self.species_fitness_func = species_fitness_func
        self.max_stagnation = max_stagnation
        self.species_elitism = species_elitism
        self.elitism = elitism
        self.survival_threshold = survival_threshold
        self.statistic_reporter = statistic_reporter
        self.create_checkpoints = create_checkpoints
        self.checkpoint_frequency = checkpoint_frequency

        # Important settings for WANN roulette selection
        self.add_conn_roulette_prob = add_conn_roulette_prob
        self.add_node_roulette_prob = add_node_roulette_prob
        self.mutate_conn_roulette_prob = mutate_conn_roulette_prob
        self.mutate_act_roulette_prob = mutate_act_roulette_prob
        self.delete_conn_roulette_prob = delete_conn_roulette_prob
        self.delete_node_roulette_prob = delete_node_roulette_prob

    @abstractmethod
    def _fitness_function(self, genomes, config):
        pass

    @abstractmethod
    def _validate_input(self, X, y):
        pass

    def _set_neat_config(self, num_inputs, num_outputs):
        _neat_config = NEATConfig(self.fitness_criterion, self.fitness_threshold, self.pop_size,
                                  self.reset_on_extinction, self.no_fitness_termination)

        _genome_config = GenomeConfig(self.activation_default, self.activation_mutate_rate, self.activation_options,
                                      self.aggregation_default, self.aggregation_mutate_rate, self.aggregation_options,
                                      self.bias_init_mean,
                                      self.bias_init_stdev, self.bias_max_value, self.bias_min_value,
                                      self.bias_mutate_power,
                                      self.bias_mutate_rate,
                                      self.bias_replace_rate, self.compatibility_disjoint_coefficient,
                                      self.compatibility_weight_coefficient,
                                      self.conn_add_prob, self.conn_delete_prob, self.enabled_default,
                                      self.enabled_mutate_rate,
                                      self.feed_forward,
                                      self.initial_connection, self.node_add_prob, self.node_delete_prob,
                                      self.num_hidden, num_inputs, num_outputs, self.response_init_mean,
                                      self.response_init_stdev, self.response_max_value, self.response_min_value,
                                      self.response_mutate_power, self.response_mutate_rate, self.response_replace_rate,
                                      self.weight_init_mean,
                                      self.weight_init_stdev, self.weight_max_value, self.weight_min_value,
                                      self.weight_mutate_power,
                                      self.weight_mutate_rate, self.weight_replace_rate, self.add_conn_roulette_prob,
                                      self.add_node_roulette_prob, self.mutate_conn_roulette_prob,
                                      self.mutate_act_roulette_prob, self.delete_conn_roulette_prob,
                                      self.delete_node_roulette_prob)

        _reproduction_config = ReproductionConfig(self.elitism, self.survival_threshold)
        _species_config = SpeciesConfig(self.compatibility_threshold)
        _stagnation_config = StagnationConfig(self.species_fitness_func, self.max_stagnation, self.species_elitism)

        _config = ConfigParser(WannGenome, WannReproduction, neat.DefaultSpeciesSet,
                               neat.DefaultStagnation, _neat_config, _genome_config, _reproduction_config,
                               _species_config, _stagnation_config)
        return _config

    def _setup_neat(self):
        p_ = neat.Population(self.config_)
        self.stats_ = neat.StatisticsReporter()
        p_.add_reporter(self.stats_)
        if self.create_checkpoints:
            p_.add_reporter(neat.Checkpointer(self.checkpoint_frequency))
        if self.statistic_reporter:
            p_.add_reporter(neat.StdOutReporter(self.statistic_reporter))
        return p_

    def _more_tags(self):
        return {'poor_score': True}

    def _fit(self, X, y):
        """
        Validate the inputs.
        Assign private NEAT config variables.
        Run the WANN algorithm.
        Displays the best genome.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_samples, n_features)
            The input data.

        y : ndarray, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification).

        Returns
        -------
        self : WANNClassifier or WANNRegressor object with the best genome (neural network)
        """
        X, y = self._validate_input(X, y)
        self.X_ = X
        self.y_ = y
        self.is_fitted_ = True
        self.pairwise_ = True
        if hasattr(self, "classes_"):
            self.config_ = self._set_neat_config(X.shape[ 1 ], len(self.classes_))
        else:
            self.config_ = self._set_neat_config(X.shape[ 1 ], 1)

        self.p_ = self._setup_neat()
        self.winner_genome_ = self.p_.run(self._fitness_function, self.number_of_generations)

        print(self.winner_genome_)

        return self

    def fit(self, X, y):
        """Fit the model to data matrix X and target(s) y.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_samples, n_features)
            The input data.

        y : ndarray, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification).

        Returns
        -------
        self : WANNClassifier or WANNRegressor object with the best genome (neural network)
        """
        return self._fit(X, y)


class WANNClassifier(BaseNEAT, ClassifierMixin):
    """WANN classifier
    """

    def __init__(self,
                 single_shared_weights=None,
                 number_of_generations=100,
                 fitness_criterion='max',
                 fitness_threshold=0.95,
                 pop_size=150,
                 reset_on_extinction=0,
                 no_fitness_termination=0,
                 activation_default='sigmoid',
                 activation_mutate_rate=1.0,
                 activation_options='sigmoid relu tanh gauss inv hat clamped sin square abs exp identity',
                 aggregation_default='sum',
                 aggregation_mutate_rate=0.0,
                 aggregation_options='sum',
                 bias_init_mean=0.0,
                 bias_init_stdev=1.0,
                 bias_max_value=30.0,
                 bias_min_value=-30.0,
                 bias_mutate_power=0.5,
                 bias_mutate_rate=0.7,
                 bias_replace_rate=0.1,
                 compatibility_disjoint_coefficient=1.0,
                 compatibility_weight_coefficient=0.5,
                 conn_add_prob=1.0,
                 conn_delete_prob=1.0,
                 enabled_default=1,
                 enabled_mutate_rate=1.0,
                 feed_forward='true',
                 initial_connection='full_direct',
                 node_add_prob=1.0,
                 node_delete_prob=1.0,
                 num_hidden=0,
                 response_init_mean=1.0,
                 response_init_stdev=0.0,
                 response_max_value=30.0,
                 response_min_value=-30.0,
                 response_mutate_power=0.0,
                 response_mutate_rate=0.0,
                 response_replace_rate=0.0,
                 weight_init_mean=0.0,
                 weight_init_stdev=1.0,
                 weight_max_value=30,
                 weight_min_value=-30,
                 weight_mutate_power=0.5,
                 weight_mutate_rate=0.8,
                 weight_replace_rate=0.1,
                 compatibility_threshold=3.0,
                 species_fitness_func='max',
                 max_stagnation=20,
                 species_elitism=2,
                 elitism=2,
                 survival_threshold=0.2,
                 statistic_reporter=1,
                 create_checkpoints=0,
                 checkpoint_frequency=20,
                 add_conn_roulette_prob=0.25,
                 add_node_roulette_prob=0.25,
                 mutate_conn_roulette_prob=0.05,
                 mutate_act_roulette_prob=0.50,
                 delete_conn_roulette_prob=0.20,
                 delete_node_roulette_prob=0.20):
        super().__init__(single_shared_weights=single_shared_weights,
                         number_of_generations=number_of_generations,
                         fitness_criterion=fitness_criterion,
                         fitness_threshold=fitness_threshold,
                         pop_size=pop_size,
                         reset_on_extinction=reset_on_extinction,
                         no_fitness_termination=no_fitness_termination,
                         activation_default=activation_default,
                         activation_mutate_rate=activation_mutate_rate,
                         activation_options=activation_options,
                         aggregation_default=aggregation_default,
                         aggregation_mutate_rate=aggregation_mutate_rate,
                         aggregation_options=aggregation_options,
                         bias_init_mean=bias_init_mean,
                         bias_init_stdev=bias_init_stdev,
                         bias_max_value=bias_max_value,
                         bias_min_value=bias_min_value,
                         bias_mutate_power=bias_mutate_power,
                         bias_mutate_rate=bias_mutate_rate,
                         bias_replace_rate=bias_replace_rate,
                         compatibility_disjoint_coefficient=compatibility_disjoint_coefficient,
                         compatibility_weight_coefficient=compatibility_weight_coefficient,
                         conn_add_prob=conn_add_prob,
                         conn_delete_prob=conn_delete_prob,
                         enabled_default=enabled_default,
                         enabled_mutate_rate=enabled_mutate_rate,
                         feed_forward=feed_forward,
                         initial_connection=initial_connection,
                         node_add_prob=node_add_prob,
                         node_delete_prob=node_delete_prob,
                         num_hidden=num_hidden,
                         response_init_mean=response_init_mean,
                         response_init_stdev=response_init_stdev,
                         response_max_value=response_max_value,
                         response_min_value=response_min_value,
                         response_mutate_power=response_mutate_power,
                         response_mutate_rate=response_mutate_rate,
                         response_replace_rate=response_replace_rate,
                         weight_init_mean=weight_init_mean,
                         weight_init_stdev=weight_init_stdev,
                         weight_max_value=weight_max_value,
                         weight_min_value=weight_min_value,
                         weight_mutate_power=weight_mutate_power,
                         weight_mutate_rate=weight_mutate_rate,
                         weight_replace_rate=weight_replace_rate,
                         compatibility_threshold=compatibility_threshold,
                         species_fitness_func=species_fitness_func,
                         max_stagnation=max_stagnation,
                         species_elitism=species_elitism,
                         elitism=elitism,
                         survival_threshold=survival_threshold,
                         statistic_reporter=statistic_reporter,
                         create_checkpoints=create_checkpoints,
                         checkpoint_frequency=checkpoint_frequency,
                         add_conn_roulette_prob=add_conn_roulette_prob,
                         add_node_roulette_prob=add_node_roulette_prob,
                         mutate_conn_roulette_prob=mutate_conn_roulette_prob,
                         mutate_act_roulette_prob=mutate_act_roulette_prob,
                         delete_conn_roulette_prob=delete_conn_roulette_prob,
                         delete_node_roulette_prob=delete_node_roulette_prob)
        if self.single_shared_weights is None:
            self.single_shared_weights = [ -2.0, -1.0, -0.5, 0.5, 1.0, 2.0 ]

    def predict(self, X, single_weight=None):
        """Predict using the NEAT classifier

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        single_weight: float
            Single shared weight value

        Returns
        -------
        y : ndarray, shape (n_samples,) or (n_samples, n_classes)
            The predicted classes.
        """
        check_is_fitted(self)
        X = check_array(X)

        # Note: We raise the ValueError for the scikit-learn unit tests to pass!
        if X.shape[ 1 ] is not self.config_.genome_config.num_inputs:
            raise ValueError

        if single_weight is not None:
            # Set the user defined single shared weight
            for connection in self.winner_genome_.connections.values():
                connection.weight = single_weight

        net = neat.nn.FeedForwardNetwork.create(self.winner_genome_, self.config_)

        predictions = np.empty(X.shape[ 0 ], ).astype(int)
        for i in range(0, X.shape[ 0 ]):
            output = net.activate(X[ i ])
            softmax_result = softmax(output)
            if len(softmax_result) > 1:
                class_output = np.argmax(((softmax_result / np.max(softmax_result)) == 1))
            else:
                class_output = self.label_encoder_.transform(softmax_result)
            predictions[ i ] = class_output

        return self.label_encoder_.inverse_transform(predictions)

    def _fitness_function(self, genomes, config):
        """Evaluation function for calculating the fitness of genomes.
        Applies the softmax function for the classification output and calculates the accuracy score.

        Parameters
        ----------
        genomes : {list} of all genomes
        config : {ConfigParser} parsed config file

        Returns
        -------
        void : Assigns the fitness to each genome
        """
        for genome_id, genome in genomes:
            reward = np.empty(len(self.single_shared_weights))
            for iVal in range(len(self.single_shared_weights)):
                # Set single-shared weight values
                for connection in genome.connections.values():
                    connection.weight = self.single_shared_weights[ iVal ]

                net = neat.nn.FeedForwardNetwork.create(genome, config)
                predictions = np.empty(self.X_.shape[ 0 ])
                for i in range(0, self.X_.shape[ 0 ]):
                    output = net.activate(self.X_[ i ])
                    softmax_result = softmax(output)
                    if len(softmax_result) > 1:
                        class_output = np.argmax(((softmax_result / np.max(softmax_result)) == 1)).astype(int)
                    else:
                        class_output = softmax_result[ 0 ]
                    predictions[ i ] = class_output

                reward[ iVal ] = accuracy_score(self.y_, predictions)

            genome.fitness = np.mean(reward)
            genome.fitMax = np.max(reward)

    def fit(self, X, y):
        """Fit the model to data matrix X and target(s) y.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_samples, n_features)
            The input data.

        y : ndarray, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification).

        Returns
        -------
        self : NEATClassifier object with the best genome (neural network)
        """
        return self._fit(X, y)

    def _validate_input(self, X, y):
        """
        Encode target labels with value between 0 and n_classes-1 with LabelEncoder.
        Get unique class labels.
        Input validation for sklearn estimators.

        Parameters
        ----------
        X : nd-array, list or sparse matrix
            Input data.

        y : nd-array, list or sparse matrix
            Labels.

        Returns
        -------
        X : object
            The converted and validated X.

        y : object
            The converted and validated y.
        """
        self.label_encoder_ = LabelEncoder()
        self.label_encoder_.fit(y)
        self.classes_ = unique_labels(y)
        y = self.label_encoder_.transform(y)
        X, y = check_X_y(X, y)
        return X, y


class WANNRegressor(BaseNEAT, RegressorMixin):
    """WANN regressor
    """

    def __init__(self,
                 single_shared_weights=None,
                 number_of_generations=100,
                 fitness_criterion='max',
                 fitness_threshold=0.95,
                 pop_size=150,
                 reset_on_extinction=0,
                 no_fitness_termination=0,
                 activation_default='relu',
                 activation_mutate_rate=0.50,
                 activation_options='sigmoid relu tanh gauss inv hat clamped sin square abs exp identity',
                 aggregation_default='sum',
                 aggregation_mutate_rate=0.0,
                 aggregation_options='sum',
                 bias_init_mean=0.0,
                 bias_init_stdev=1.0,
                 bias_max_value=30.0,
                 bias_min_value=-30.0,
                 bias_mutate_power=0.5,
                 bias_mutate_rate=0.7,
                 bias_replace_rate=0.1,
                 compatibility_disjoint_coefficient=1.0,
                 compatibility_weight_coefficient=0.5,
                 conn_add_prob=0.50,
                 conn_delete_prob=0.50,
                 enabled_default=1,
                 enabled_mutate_rate=0.05,
                 feed_forward='true',
                 initial_connection='full_direct',
                 node_add_prob=0.25,
                 node_delete_prob=0.25,
                 num_hidden=0,
                 response_init_mean=1.0,
                 response_init_stdev=0.0,
                 response_max_value=30.0,
                 response_min_value=-30.0,
                 response_mutate_power=0.0,
                 response_mutate_rate=0.0,
                 response_replace_rate=0.0,
                 weight_init_mean=0.0,
                 weight_init_stdev=1.0,
                 weight_max_value=30,
                 weight_min_value=-30,
                 weight_mutate_power=0.5,
                 weight_mutate_rate=0.8,
                 weight_replace_rate=0.1,
                 compatibility_threshold=3.0,
                 species_fitness_func='max',
                 max_stagnation=20,
                 species_elitism=2,
                 elitism=2,
                 survival_threshold=0.2,
                 statistic_reporter=1,
                 create_checkpoints=0,
                 checkpoint_frequency=20,
                 add_conn_roulette_prob=0.50,
                 add_node_roulette_prob=0.50,
                 mutate_conn_roulette_prob=0.05,
                 mutate_act_roulette_prob=0.50,
                 delete_conn_roulette_prob=0.20,
                 delete_node_roulette_prob=0.20):
        super().__init__(single_shared_weights=single_shared_weights,
                         number_of_generations=number_of_generations,
                         fitness_criterion=fitness_criterion,
                         fitness_threshold=fitness_threshold,
                         pop_size=pop_size,
                         reset_on_extinction=reset_on_extinction,
                         no_fitness_termination=no_fitness_termination,
                         activation_default=activation_default,
                         activation_mutate_rate=activation_mutate_rate,
                         activation_options=activation_options,
                         aggregation_default=aggregation_default,
                         aggregation_mutate_rate=aggregation_mutate_rate,
                         aggregation_options=aggregation_options,
                         bias_init_mean=bias_init_mean,
                         bias_init_stdev=bias_init_stdev,
                         bias_max_value=bias_max_value,
                         bias_min_value=bias_min_value,
                         bias_mutate_power=bias_mutate_power,
                         bias_mutate_rate=bias_mutate_rate,
                         bias_replace_rate=bias_replace_rate,
                         compatibility_disjoint_coefficient=compatibility_disjoint_coefficient,
                         compatibility_weight_coefficient=compatibility_weight_coefficient,
                         conn_add_prob=conn_add_prob,
                         conn_delete_prob=conn_delete_prob,
                         enabled_default=enabled_default,
                         enabled_mutate_rate=enabled_mutate_rate,
                         feed_forward=feed_forward,
                         initial_connection=initial_connection,
                         node_add_prob=node_add_prob,
                         node_delete_prob=node_delete_prob,
                         num_hidden=num_hidden,
                         response_init_mean=response_init_mean,
                         response_init_stdev=response_init_stdev,
                         response_max_value=response_max_value,
                         response_min_value=response_min_value,
                         response_mutate_power=response_mutate_power,
                         response_mutate_rate=response_mutate_rate,
                         response_replace_rate=response_replace_rate,
                         weight_init_mean=weight_init_mean,
                         weight_init_stdev=weight_init_stdev,
                         weight_max_value=weight_max_value,
                         weight_min_value=weight_min_value,
                         weight_mutate_power=weight_mutate_power,
                         weight_mutate_rate=weight_mutate_rate,
                         weight_replace_rate=weight_replace_rate,
                         compatibility_threshold=compatibility_threshold,
                         species_fitness_func=species_fitness_func,
                         max_stagnation=max_stagnation,
                         species_elitism=species_elitism,
                         elitism=elitism,
                         survival_threshold=survival_threshold,
                         statistic_reporter=statistic_reporter,
                         create_checkpoints=create_checkpoints,
                         checkpoint_frequency=checkpoint_frequency,
                         add_conn_roulette_prob=add_conn_roulette_prob,
                         add_node_roulette_prob=add_node_roulette_prob,
                         mutate_conn_roulette_prob=mutate_conn_roulette_prob,
                         mutate_act_roulette_prob=mutate_act_roulette_prob,
                         delete_conn_roulette_prob=delete_conn_roulette_prob,
                         delete_node_roulette_prob=delete_node_roulette_prob)
        if self.single_shared_weights is None:
            self.single_shared_weights = [ -2.0, -1.0, -0.5, 0.5, 1.0, 2.0 ]

    def predict(self, X):
        """Predict using the NEATRegressor model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        Returns
        -------
        y : ndarray of shape (n_samples, n_outputs)
            The predicted values.
        """
        check_is_fitted(self, [ 'X_', 'y_' ])
        X = check_array(X)
        net = neat.nn.FeedForwardNetwork.create(self.winner_genome_, self.config_)
        predictions = np.empty(X.shape[ 0 ], )
        for i in range(0, X.shape[ 0 ]):
            output = net.activate(X[ i ])
            predictions[ i ] = output[ 0 ]

        return predictions

    def _fitness_function(self, genomes, config):
        """Evaluation function for calculating the fitness of genomes. Calculates the r2_score

        Parameters
        ----------
        genomes : {list} of all genomes
        config : {ConfigParser} parsed config file

        Returns
        -------
        void : Assigns the fitness to each genome
        """
        for genome_id, genome in genomes:
            reward = np.empty(len(self.single_shared_weights))
            for iVal in range(len(self.single_shared_weights)):
                # Set single-shared weight values
                for connection in genome.connections.values():
                    connection.weight = self.single_shared_weights[ iVal ]
                net = neat.nn.FeedForwardNetwork.create(genome, config)
                predictions = np.empty(self.y_.shape[ 0 ])
                for i in range(0, self.y_.shape[ 0 ]):
                    output = net.activate(self.X_[ i ])
                    predictions[ i ] = output[ 0 ]

                reward[ iVal ] = r2_score(self.y_, predictions)

            genome.fitness = np.mean(reward)
            genome.fitMax = np.max(reward)

    def _validate_input(self, X, y):
        """Input validation for sklearn estimators

        Parameters
        ----------
        X : nd-array, list or sparse matrix
            Input data.

        y : nd-array, list or sparse matrix
            Labels.

        Returns
        -------
        X : object
            The converted and validated X.

        y : object
            The converted and validated y.
        """
        X, y = check_X_y(X, y)
        return X, y



from ._neat import NEATClassifier
from ._neat import NEATRegressor
from ._wann import WANNClassifier
from ._wann import WANNRegressor


from ._version import __version__

__all__ = ['NEATClassifier', 'NEATRegressor', 'WANNClassifier', 'WANNRegressor',
           '__version__']



"""
Adapted from https://github.com/google/brain-tokyo-workshop/blob/master/WANNRelease/WANN/wann_src/nsga_sort.py
"""

import numpy as np
import warnings


def nsga_sort(objVals, returnFronts=False):
    """Returns ranking of objective values based on non-dominated sorting.
    Optionally returns fronts (useful for visualization).
    NOTE: Assumes maximization of objective function
    Args:
      objVals - (np_array) - Objective values of each individual
                [nInds X nObjectives]
    Returns:
      rank    - (np_array) - Rank in population of each individual
              int([nIndividuals X 1])
      front   - (np_array) - Pareto front of each individual
              int([nIndividuals X 1])
    Todo:
      * Extend to N objectives
    """
    # Sort by dominance into fronts
    fronts = getFronts(objVals)

    # Rank each front by crowding distance
    for f in range(len(fronts)):
        x1 = objVals[ fronts[ f ], 0 ]
        x2 = objVals[ fronts[ f ], 1 ]
        crowdDist = getCrowdingDist(x1) + getCrowdingDist(x2)
        frontRank = np.argsort(-crowdDist)
        fronts[ f ] = [ fronts[ f ][ i ] for i in frontRank ]

    # Convert to ranking
    tmp = [ ind for front in fronts for ind in front ]
    rank = np.empty_like(tmp)
    rank[ tmp ] = np.arange(len(tmp))

    if returnFronts is True:
        return rank, fronts
    else:
        return rank

def getFronts(objVals):
    """Fast non-dominated sort.
    Args:
      objVals - (np_array) - Objective values of each individual
                [nInds X nObjectives]
    Returns:
      front   - [list of lists] - One list for each front:
                                  list of indices of individuals in front
    Todo:
      * Extend to N objectives
    [adapted from: https://github.com/haris989/NSGA-II]
    """

    values1 = objVals[ :, 0 ]
    values2 = objVals[ :, 1 ]

    S = [ [ ] for i in range(0, len(values1)) ]
    front = [ [ ] ]
    n = [ 0 for i in range(0, len(values1)) ]
    rank = [ 0 for i in range(0, len(values1)) ]
    # Get domination relations
    for p in range(0, len(values1)):
        S[ p ] = [ ]
        n[ p ] = 0
        for q in range(0, len(values1)):
            if (values1[ p ] > values1[ q ] and values2[ p ] > values2[ q ]) \
                    or (values1[ p ] >= values1[ q ] and values2[ p ] > values2[ q ]) \
                    or (values1[ p ] > values1[ q ] and values2[ p ] >= values2[ q ]):
                if q not in S[ p ]:
                    S[ p ].append(q)
            elif (values1[ q ] > values1[ p ] and values2[ q ] > values2[ p ]) \
                    or (values1[ q ] >= values1[ p ] and values2[ q ] > values2[ p ]) \
                    or (values1[ q ] > values1[ p ] and values2[ q ] >= values2[ p ]):
                n[ p ] = n[ p ] + 1
        if n[ p ] == 0:
            rank[ p ] = 0
            if p not in front[ 0 ]:
                front[ 0 ].append(p)

    # Assign fronts
    i = 0
    while (front[ i ] != [ ]):
        Q = [ ]
        for p in front[ i ]:
            for q in S[ p ]:
                n[ q ] = n[ q ] - 1
                if (n[ q ] == 0):
                    rank[ q ] = i + 1
                    if q not in Q:
                        Q.append(q)
        i = i + 1
        front.append(Q)
    del front[ len(front) - 1 ]
    return front


def getCrowdingDist(objVector):
    """Returns crowding distance of a vector of values, used once on each front.
    Note: Crowding distance of individuals at each end of front is infinite, as they don't have a neighbor.
    Args:
      objVector - (np_array) - Objective values of each individual
                  [nInds X nObjectives]
    Returns:
      dist      - (np_array) - Crowding distance of each individual
                  [nIndividuals X 1]
    """
    # Order by objective value
    key = np.argsort(objVector)
    sortedObj = objVector[ key ]

    # Distance from values on either side
    shiftVec = np.r_[ np.inf, sortedObj, np.inf ]  # Edges have infinite distance
    warnings.filterwarnings("ignore", category=RuntimeWarning)  # inf on purpose
    prevDist = np.abs(sortedObj - shiftVec[ :-2 ])
    nextDist = np.abs(sortedObj - shiftVec[ 2: ])
    crowd = prevDist + nextDist
    if (sortedObj[ -1 ] - sortedObj[ 0 ]) > 0:
        crowd *= abs((1 / sortedObj[ -1 ] - sortedObj[ 0 ]))  # Normalize by fitness range

    # Restore original order
    dist = np.empty(len(key))
    dist[ key ] = crowd[ : ]

    return dist


class NEATConfig:
    def __init__(self, fitness_criterion, fitness_threshold, pop_size,
                 reset_on_extinction, no_fitness_termination):
        self.fitness_criterion = fitness_criterion
        self.fitness_threshold = fitness_threshold
        self.pop_size = pop_size
        self.reset_on_extinction = reset_on_extinction
        self.no_fitness_termination = no_fitness_termination


class GenomeConfig:
    def __init__(self, activation_default, activation_mutate_rate, activation_options,
                 aggregation_default, aggregation_mutate_rate, aggregation_options, bias_init_mean,
                 bias_init_stdev, bias_max_value, bias_min_value, bias_mutate_power, bias_mutate_rate,
                 bias_replace_rate, compatibility_disjoint_coefficient, compatibility_weight_coefficient,
                 conn_add_prob, conn_delete_prob, enabled_default, enabled_mutate_rate, feed_forward,
                 initial_connection, node_add_prob, node_delete_prob, num_hidden, num_inputs, num_outputs,
                 response_init_mean, response_init_stdev, response_max_value, response_min_value,
                 response_mutate_power, response_mutate_rate, response_replace_rate, weight_init_mean,
                 weight_init_stdev, weight_max_value, weight_min_value, weight_mutate_power,
                 weight_mutate_rate, weight_replace_rate, add_conn_roulette_prob, add_node_roulette_prob,
                 mutate_conn_roulette_prob, mutate_act_roulette_prob, delete_conn_roulette_prob,
                 delete_node_roulette_prob):

        self.activation_default = activation_default
        self.activation_mutate_rate = activation_mutate_rate
        self.activation_options = activation_options
        self.aggregation_default = aggregation_default
        self.aggregation_mutate_rate = aggregation_mutate_rate
        self.aggregation_options = aggregation_options
        self.bias_init_mean = bias_init_mean
        self.bias_init_stdev = bias_init_stdev
        self.bias_max_value = bias_max_value
        self.bias_min_value = bias_min_value
        self.bias_mutate_power = bias_mutate_power
        self.bias_mutate_rate = bias_mutate_rate
        self.bias_replace_rate = bias_replace_rate
        self.compatibility_disjoint_coefficient = compatibility_disjoint_coefficient
        self.compatibility_weight_coefficient = compatibility_weight_coefficient
        self.conn_add_prob = conn_add_prob
        self.conn_delete_prob = conn_delete_prob
        self.enabled_default = enabled_default
        self.enabled_mutate_rate = enabled_mutate_rate
        self.feed_forward = feed_forward
        self.initial_connection = initial_connection
        self.node_add_prob = node_add_prob
        self.node_delete_prob = node_delete_prob
        self.num_hidden = num_hidden
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs
        self.response_init_mean = response_init_mean
        self.response_init_stdev = response_init_stdev
        self.response_max_value = response_max_value
        self.response_min_value = response_min_value
        self.response_mutate_power = response_mutate_power
        self.response_mutate_rate = response_mutate_rate
        self.response_replace_rate = response_replace_rate
        self.weight_init_mean = weight_init_mean
        self.weight_init_stdev = weight_init_stdev
        self.weight_max_value = weight_max_value
        self.weight_min_value = weight_min_value
        self.weight_mutate_power = weight_mutate_power
        self.weight_mutate_rate = weight_mutate_rate
        self.weight_replace_rate = weight_replace_rate
        self.add_conn_roulette_prob = add_conn_roulette_prob
        self.add_node_roulette_prob = add_node_roulette_prob
        self.mutate_conn_roulette_prob = mutate_conn_roulette_prob
        self.mutate_act_roulette_prob = mutate_act_roulette_prob
        self.delete_conn_roulette_prob = delete_conn_roulette_prob
        self.delete_node_roulette_prob = delete_node_roulette_prob


class ReproductionConfig:
    def __init__(self, elitism, survival_threshold):
        self.elitism = elitism
        self.survival_threshold = survival_threshold


class StagnationConfig:
    def __init__(self, species_fitness_func, max_stagnation, species_elitism):
        self.species_fitness_func = species_fitness_func
        self.max_stagnation = max_stagnation
        self.species_elitism = species_elitism


class SpeciesConfig:
    def __init__(self, compatibility_threshold):
        self.compatibility_threshold = compatibility_threshold


class ConfigParser:
    def __init__(self, genome_type, reproduction_type, species_set_type, stagnation_type, neat_config, genome_config,
                 reproduction_config, species_config, stagnation_config):
        self.genome_type = genome_type
        self.reproduction_type = reproduction_type
        self.species_set_type = species_set_type
        self.stagnation_type = stagnation_type

        # Assign NEAT attributes
        for attr, value in neat_config.__dict__.items():
            setattr(self, attr, value)

        # Assign default Genome attributes
        genome_config_dict = {}
        for attr, value in genome_config.__dict__.items():
            genome_config_dict.update({attr: value})

        # Assign default Reproduction attributes
        reproduction_config_dict = {}
        for attr, value in reproduction_config.__dict__.items():
            reproduction_config_dict.update({attr: value})

        # # Assign default Species attributes
        species_config_dict = {}
        for attr, value in species_config.__dict__.items():
            species_config_dict.update({attr: value})

        # Assign default Stagnation attributes
        stagnation_config_dict = {}
        for attr, value in stagnation_config.__dict__.items():
            stagnation_config_dict.update({attr: value})

        self.genome_config = self.genome_type.parse_config(genome_config_dict)
        self.reproduction_config = self.reproduction_type.parse_config(reproduction_config_dict)
        self.species_set_config = self.species_set_type.parse_config(species_config_dict)
        self.stagnation_config = self.stagnation_type.parse_config(stagnation_config_dict)


"""Sklearn NEAT
"""

# Author: Marko Mlakar <markomlakar2@gmail.com>
# License: BSD-3-Clause License

from abc import ABCMeta, abstractmethod
from sklearn.metrics import accuracy_score, r2_score
import neat
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.multiclass import unique_labels
import multiprocessing
from ._config import NEATConfig, ConfigParser, GenomeConfig, ReproductionConfig, StagnationConfig, SpeciesConfig
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted


def softmax(x):
    """Compute softmax values for each sets of scores in x"""
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)

class BaseNEAT(BaseEstimator, metaclass=ABCMeta):
    """Base class for NEAT classification and regression.

    Warning: This class should not be used directly. Use derived classes instead!

    """

    @abstractmethod
    def __init__(self, number_of_generations, fitness_criterion, fitness_threshold, pop_size,
                 reset_on_extinction, no_fitness_termination, activation_default, activation_mutate_rate,
                 activation_options,
                 aggregation_default, aggregation_mutate_rate, aggregation_options, bias_init_mean,
                 bias_init_stdev, bias_max_value, bias_min_value, bias_mutate_power, bias_mutate_rate,
                 bias_replace_rate, compatibility_disjoint_coefficient, compatibility_weight_coefficient,
                 conn_add_prob, conn_delete_prob, enabled_default, enabled_mutate_rate, feed_forward,
                 initial_connection, node_add_prob, node_delete_prob, num_hidden,
                 response_init_mean, response_init_stdev, response_max_value, response_min_value,
                 response_mutate_power, response_mutate_rate, response_replace_rate, weight_init_mean,
                 weight_init_stdev, weight_max_value, weight_min_value, weight_mutate_power,
                 weight_mutate_rate, weight_replace_rate, compatibility_threshold, species_fitness_func,
                 max_stagnation, species_elitism, elitism, survival_threshold, statistic_reporter,
                 create_checkpoints, checkpoint_frequency):
        self.number_of_generations = number_of_generations
        self.fitness_criterion = fitness_criterion
        self.fitness_threshold = fitness_threshold
        self.pop_size = pop_size
        self.reset_on_extinction = reset_on_extinction
        self.no_fitness_termination = no_fitness_termination
        self.activation_default = activation_default
        self.activation_mutate_rate = activation_mutate_rate
        self.activation_options = activation_options
        self.aggregation_default = aggregation_default
        self.aggregation_mutate_rate = aggregation_mutate_rate
        self.aggregation_options = aggregation_options
        self.bias_init_mean = bias_init_mean
        self.bias_init_stdev = bias_init_stdev
        self.bias_max_value = bias_max_value
        self.bias_min_value = bias_min_value
        self.bias_mutate_power = bias_mutate_power
        self.bias_mutate_rate = bias_mutate_rate
        self.bias_replace_rate = bias_replace_rate
        self.compatibility_disjoint_coefficient = compatibility_disjoint_coefficient
        self.compatibility_weight_coefficient = compatibility_weight_coefficient
        self.conn_add_prob = conn_add_prob
        self.conn_delete_prob = conn_delete_prob
        self.enabled_default = enabled_default
        self.enabled_mutate_rate = enabled_mutate_rate
        self.feed_forward = feed_forward
        self.initial_connection = initial_connection
        self.node_add_prob = node_add_prob
        self.node_delete_prob = node_delete_prob
        self.num_hidden = num_hidden
        self.response_init_mean = response_init_mean
        self.response_init_stdev = response_init_stdev
        self.response_max_value = response_max_value
        self.response_min_value = response_min_value
        self.response_mutate_power = response_mutate_power
        self.response_mutate_rate = response_mutate_rate
        self.response_replace_rate = response_replace_rate
        self.weight_init_mean = weight_init_mean
        self.weight_init_stdev = weight_init_stdev
        self.weight_max_value = weight_max_value
        self.weight_min_value = weight_min_value
        self.weight_mutate_power = weight_mutate_power
        self.weight_mutate_rate = weight_mutate_rate
        self.weight_replace_rate = weight_replace_rate
        self.compatibility_threshold = compatibility_threshold
        self.species_fitness_func = species_fitness_func
        self.max_stagnation = max_stagnation
        self.species_elitism = species_elitism
        self.elitism = elitism
        self.survival_threshold = survival_threshold
        self.statistic_reporter = statistic_reporter
        self.create_checkpoints = create_checkpoints
        self.checkpoint_frequency = checkpoint_frequency

    @abstractmethod
    def _fitness_function(self, genomes, config):
        pass

    @abstractmethod
    def _validate_input(self, X, y):
        pass

    def _set_neat_config(self, num_inputs, num_outputs):
        _neat_config = NEATConfig(self.fitness_criterion, self.fitness_threshold, self.pop_size,
                                  self.reset_on_extinction, self.no_fitness_termination)

        _genome_config = GenomeConfig(self.activation_default, self.activation_mutate_rate, self.activation_options,
                                      self.aggregation_default, self.aggregation_mutate_rate, self.aggregation_options,
                                      self.bias_init_mean,
                                      self.bias_init_stdev, self.bias_max_value, self.bias_min_value,
                                      self.bias_mutate_power,
                                      self.bias_mutate_rate,
                                      self.bias_replace_rate, self.compatibility_disjoint_coefficient,
                                      self.compatibility_weight_coefficient,
                                      self.conn_add_prob, self.conn_delete_prob, self.enabled_default,
                                      self.enabled_mutate_rate,
                                      self.feed_forward,
                                      self.initial_connection, self.node_add_prob, self.node_delete_prob,
                                      self.num_hidden, num_inputs,
                                      num_outputs,
                                      self.response_init_mean, self.response_init_stdev, self.response_max_value,
                                      self.response_min_value,
                                      self.response_mutate_power, self.response_mutate_rate, self.response_replace_rate,
                                      self.weight_init_mean,
                                      self.weight_init_stdev, self.weight_max_value, self.weight_min_value,
                                      self.weight_mutate_power,
                                      self.weight_mutate_rate, self.weight_replace_rate,
                                      # TODO: Refactor this!!
                                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

        _reproduction_config = ReproductionConfig(self.elitism, self.survival_threshold)
        _species_config = SpeciesConfig(self.compatibility_threshold)
        _stagnation_config = StagnationConfig(self.species_fitness_func, self.max_stagnation, self.species_elitism)

        _config = ConfigParser(neat.DefaultGenome, neat.DefaultReproduction, neat.DefaultSpeciesSet,
                               neat.DefaultStagnation, _neat_config, _genome_config, _reproduction_config,
                               _species_config, _stagnation_config)
        return _config

    def _setup_neat(self):
        p_ = neat.Population(self.config_)
        self.stats_ = neat.StatisticsReporter()
        p_.add_reporter(self.stats_)
        if self.statistic_reporter:
            p_.add_reporter(neat.StdOutReporter(self.statistic_reporter))
        if self.create_checkpoints:
            p_.add_reporter(neat.Checkpointer(self.checkpoint_frequency))
        return p_

    def _more_tags(self):
        return {'poor_score': True}

    def _fit(self, X, y):
        X, y = self._validate_input(X,y)
        self.X_ = X
        self.y_ = y
        self.is_fitted_ = True
        self.pairwise_ = True
        if hasattr(self, "classes_"):
            self.config_ = self._set_neat_config(X.shape[1], len(self.classes_))
        else:
            self.config_ = self._set_neat_config(X.shape[1], 1)

        self.p_ = self._setup_neat()
        self.winner_genome_ = self.p_.run(self._fitness_function, self.number_of_generations)

        print(self.winner_genome_)

        return self


    def fit(self, X, y):
        """Fit the model to data matrix X and target(s) y.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_samples, n_features)
            The input data.

        y : ndarray, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification).

        Returns
        -------
        self : NEATClassifier or NEATRegressor object with the best genome (neural network)
        """
        return self._fit(X, y)




class NEATClassifier(BaseNEAT, ClassifierMixin):
    """NEAT classifier
    """

    # TODO: Add model description

    def __init__(self, number_of_generations=100,
                 fitness_criterion='max',
                 fitness_threshold=0.95,
                 pop_size=150,
                 reset_on_extinction=0,
                 no_fitness_termination=0,
                 activation_default='sigmoid',
                 activation_mutate_rate=0.00,
                 activation_options='sigmoid relu tanh gauss inv hat clamped sin square abs exp identity',
                 aggregation_default='sum',
                 aggregation_mutate_rate=0.0,
                 aggregation_options='sum',
                 bias_init_mean=0.0,
                 bias_init_stdev=1.0,
                 bias_max_value=30.0,
                 bias_min_value=-30.0,
                 bias_mutate_power=0.5,
                 bias_mutate_rate=0.7,
                 bias_replace_rate=0.1,
                 compatibility_disjoint_coefficient=1.0,
                 compatibility_weight_coefficient=0.5,
                 conn_add_prob=0.50,
                 conn_delete_prob=0.50,
                 enabled_default=1,
                 enabled_mutate_rate=0.01,
                 feed_forward='true',
                 initial_connection='full_direct',
                 node_add_prob=0.25,
                 node_delete_prob=0.25,
                 num_hidden=0,
                 response_init_mean=1.0,
                 response_init_stdev=0.0,
                 response_max_value=30.0,
                 response_min_value=-30.0,
                 response_mutate_power=0.0,
                 response_mutate_rate=0.0,
                 response_replace_rate=0.0,
                 weight_init_mean=0.0,
                 weight_init_stdev=1.0,
                 weight_max_value=30,
                 weight_min_value=-30,
                 weight_mutate_power=0.5,
                 weight_mutate_rate=0.8,
                 weight_replace_rate=0.1,
                 compatibility_threshold=3.0,
                 species_fitness_func='max',
                 max_stagnation=20,
                 species_elitism=2,
                 elitism=2,
                 survival_threshold=0.2,
                 statistic_reporter=1,
                 create_checkpoints=0,
                 checkpoint_frequency=20):
        super().__init__(number_of_generations=number_of_generations,
                         fitness_criterion=fitness_criterion,
                         fitness_threshold=fitness_threshold,
                         pop_size=pop_size,
                         reset_on_extinction=reset_on_extinction,
                         no_fitness_termination=no_fitness_termination,
                         activation_default=activation_default,
                         activation_mutate_rate=activation_mutate_rate,
                         activation_options=activation_options,
                         aggregation_default=aggregation_default,
                         aggregation_mutate_rate=aggregation_mutate_rate,
                         aggregation_options=aggregation_options,
                         bias_init_mean=bias_init_mean,
                         bias_init_stdev=bias_init_stdev,
                         bias_max_value=bias_max_value,
                         bias_min_value=bias_min_value,
                         bias_mutate_power=bias_mutate_power,
                         bias_mutate_rate=bias_mutate_rate,
                         bias_replace_rate=bias_replace_rate,
                         compatibility_disjoint_coefficient=compatibility_disjoint_coefficient,
                         compatibility_weight_coefficient=compatibility_weight_coefficient,
                         conn_add_prob=conn_add_prob,
                         conn_delete_prob=conn_delete_prob,
                         enabled_default=enabled_default,
                         enabled_mutate_rate=enabled_mutate_rate,
                         feed_forward=feed_forward,
                         initial_connection=initial_connection,
                         node_add_prob=node_add_prob,
                         node_delete_prob=node_delete_prob,
                         num_hidden=num_hidden,
                         response_init_mean=response_init_mean,
                         response_init_stdev=response_init_stdev,
                         response_max_value=response_max_value,
                         response_min_value=response_min_value,
                         response_mutate_power=response_mutate_power,
                         response_mutate_rate=response_mutate_rate,
                         response_replace_rate=response_replace_rate,
                         weight_init_mean=weight_init_mean,
                         weight_init_stdev=weight_init_stdev,
                         weight_max_value=weight_max_value,
                         weight_min_value=weight_min_value,
                         weight_mutate_power=weight_mutate_power,
                         weight_mutate_rate=weight_mutate_rate,
                         weight_replace_rate=weight_replace_rate,
                         compatibility_threshold=compatibility_threshold,
                         species_fitness_func=species_fitness_func,
                         max_stagnation=max_stagnation,
                         species_elitism=species_elitism,
                         elitism=elitism,
                         survival_threshold=survival_threshold,
                         statistic_reporter=statistic_reporter,
                         create_checkpoints=create_checkpoints,
                         checkpoint_frequency=checkpoint_frequency)

    def predict(self, X):
        """Predict using the NEAT classifier

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        Returns
        -------
        y : ndarray, shape (n_samples,) or (n_samples, n_classes)
            The predicted classes.
        """
        check_is_fitted(self)
        X = check_array(X)

        if X.shape[1] is not self.config_.genome_config.num_inputs:
            raise ValueError

        net = neat.nn.FeedForwardNetwork.create(self.winner_genome_, self.config_)

        predictions = np.empty(X.shape[0], ).astype(int)
        for i in range(0, X.shape[0]):
            output = net.activate(X[i])
            softmax_result = softmax(output)
            if len(softmax_result) > 1:
                class_output = np.argmax(((softmax_result / np.max(softmax_result)) == 1))
            else:
                class_output = self.label_encoder_.transform(softmax_result)
            predictions[i] = class_output

        return self.label_encoder_.inverse_transform(predictions)

    def _fitness_function(self, genomes, config):
        """Evaluation function for calculating the fitness of genomes.
        Applies the softmax function for the classification output and calculates the accuracy score.

        Parameters
        ----------
        genomes : {list} of all genomes
        config : {ConfigParser} parsed config file

        Returns
        -------
        void : Assigns the fitness to each genome
        """
        for genome_id, genome in genomes:
            net = neat.nn.FeedForwardNetwork.create(genome, config)
            predictions = np.empty(self.X_.shape[0])
            for i in range(0, self.X_.shape[0]):
                output = net.activate(self.X_[i])
                softmax_result = softmax(output)
                if len(softmax_result) > 1:
                    class_output = np.argmax(((softmax_result / np.max(softmax_result)) == 1))
                else:
                    class_output = softmax_result[0]
                predictions[i] = class_output

            genome.fitness = accuracy_score(self.y_, predictions)

    def fit(self, X, y):
        """Fit the model to data matrix X and target(s) y.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_samples, n_features)
            The input data.

        y : ndarray, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification).

        Returns
        -------
        self : NEATClassifier object with the best genome (neural network)
        """
        return self._fit(X, y)

    def _validate_input(self, X, y):
        self.label_encoder_ = LabelEncoder()
        self.label_encoder_.fit(y)
        self.classes_ = unique_labels(y)
        y = self.label_encoder_.transform(y)
        X, y = check_X_y(X, y)
        return X, y

class NEATRegressor(BaseNEAT, RegressorMixin):
    """NEAT regressor
    """
    def __init__(self, number_of_generations=100,
                 fitness_criterion='max',
                 fitness_threshold=0.95,
                 pop_size=150,
                 reset_on_extinction=0,
                 no_fitness_termination=0,
                 activation_default='relu',
                 activation_mutate_rate=0.0,
                 activation_options='sigmoid relu tanh gauss inv hat clamped sin square abs exp identity',
                 aggregation_default='sum',
                 aggregation_mutate_rate=0.0,
                 aggregation_options='sum',
                 bias_init_mean=0.0,
                 bias_init_stdev=1.0,
                 bias_max_value=30.0,
                 bias_min_value=-30.0,
                 bias_mutate_power=0.5,
                 bias_mutate_rate=0.7,
                 bias_replace_rate=0.1,
                 compatibility_disjoint_coefficient=1.0,
                 compatibility_weight_coefficient=0.5,
                 conn_add_prob=0.50,
                 conn_delete_prob=0.50,
                 enabled_default='true',
                 enabled_mutate_rate=0.01,
                 feed_forward='true',
                 initial_connection='full',
                 node_add_prob=0.25,
                 node_delete_prob=0.25,
                 num_hidden=0,
                 response_init_mean=1.0,
                 response_init_stdev=0.0,
                 response_max_value=30.0,
                 response_min_value=-30.0,
                 response_mutate_power=0.0,
                 response_mutate_rate=0.0,
                 response_replace_rate=0.0,
                 weight_init_mean=0.0,
                 weight_init_stdev=1.0,
                 weight_max_value=30,
                 weight_min_value=-30,
                 weight_mutate_power=0.5,
                 weight_mutate_rate=0.8,
                 weight_replace_rate=0.1,
                 compatibility_threshold=3.0,
                 species_fitness_func='max',
                 max_stagnation=20,
                 species_elitism=1,
                 elitism=2,
                 survival_threshold=0.2,
                 statistic_reporter=1,
                 create_checkpoints=0,
                 checkpoint_frequency=20):
        super().__init__(number_of_generations=number_of_generations,
                         fitness_criterion=fitness_criterion,
                         fitness_threshold=fitness_threshold,
                         pop_size=pop_size,
                         reset_on_extinction=reset_on_extinction,
                         no_fitness_termination=no_fitness_termination,
                         activation_default=activation_default,
                         activation_mutate_rate=activation_mutate_rate,
                         activation_options=activation_options,
                         aggregation_default=aggregation_default,
                         aggregation_mutate_rate=aggregation_mutate_rate,
                         aggregation_options=aggregation_options,
                         bias_init_mean=bias_init_mean,
                         bias_init_stdev=bias_init_stdev,
                         bias_max_value=bias_max_value,
                         bias_min_value=bias_min_value,
                         bias_mutate_power=bias_mutate_power,
                         bias_mutate_rate=bias_mutate_rate,
                         bias_replace_rate=bias_replace_rate,
                         compatibility_disjoint_coefficient=compatibility_disjoint_coefficient,
                         compatibility_weight_coefficient=compatibility_weight_coefficient,
                         conn_add_prob=conn_add_prob,
                         conn_delete_prob=conn_delete_prob,
                         enabled_default=enabled_default,
                         enabled_mutate_rate=enabled_mutate_rate,
                         feed_forward=feed_forward,
                         initial_connection=initial_connection,
                         node_add_prob=node_add_prob,
                         node_delete_prob=node_delete_prob,
                         num_hidden=num_hidden,
                         response_init_mean=response_init_mean,
                         response_init_stdev=response_init_stdev,
                         response_max_value=response_max_value,
                         response_min_value=response_min_value,
                         response_mutate_power=response_mutate_power,
                         response_mutate_rate=response_mutate_rate,
                         response_replace_rate=response_replace_rate,
                         weight_init_mean=weight_init_mean,
                         weight_init_stdev=weight_init_stdev,
                         weight_max_value=weight_max_value,
                         weight_min_value=weight_min_value,
                         weight_mutate_power=weight_mutate_power,
                         weight_mutate_rate=weight_mutate_rate,
                         weight_replace_rate=weight_replace_rate,
                         compatibility_threshold=compatibility_threshold,
                         species_fitness_func=species_fitness_func,
                         max_stagnation=max_stagnation,
                         species_elitism=species_elitism,
                         elitism=elitism,
                         survival_threshold=survival_threshold,
                         statistic_reporter=statistic_reporter,
                         create_checkpoints=create_checkpoints,
                         checkpoint_frequency=checkpoint_frequency)

    def predict(self, X):
        """Predict using the NEATRegressor model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        Returns
        -------
        y : ndarray of shape (n_samples, n_outputs)
            The predicted values.
        """
        check_is_fitted(self, [ 'X_', 'y_' ])
        X = check_array(X)
        net = neat.nn.FeedForwardNetwork.create(self.winner_genome_, self.config_)
        predictions = np.empty(X.shape[ 0 ], )
        for i in range(0, X.shape[ 0 ]):
            output = net.activate(X[ i ])
            predictions[ i ] = output[ 0 ]

        return predictions

    def _fitness_function(self, genomes, config):
        """Evaluation function for calculating the fitness of genomes. Calculates the r2_score

        Parameters
        ----------
        genomes : {list} of all genomes
        config : {ConfigParser} parsed config file

        Returns
        -------
        void : Assigns the fitness to each genome
        """
        for genome_id, genome in genomes:
            net = neat.nn.FeedForwardNetwork.create(genome, config)
            predictions = np.empty(self.y_.shape[0])
            for i in range(0, self.y_.shape[0]):
                output = net.activate(self.X_[i])
                predictions[i] = output[0]
            genome.fitness = r2_score(self.y_, predictions)

    def _validate_input(self, X, y):
        X, y = check_X_y(X, y)
        return X, y



# NOTE: If some tests fail (or take too long)
# you need to tweak the WANNClassifier() and WANNRegressor() init parameters in _wann.py!

from sklearn.utils.estimator_checks import parametrize_with_checks
from neuro_evolution import WANNClassifier
from neuro_evolution import WANNRegressor


@parametrize_with_checks([WANNRegressor(), WANNClassifier()])
def test_sklearn_compatible_estimator(estimator, check):
    check(estimator)


import neat
from neat.activations import ActivationFunctionSet
from neat.aggregations import AggregationFunctionSet
from neat.config import ConfigParameter, write_pretty_params
from neat.genes import DefaultNodeGene, DefaultConnectionGene
from neat.genome import DefaultGenomeConfig
import numpy as np
import random
from itertools import count
from neat.six_util import iteritems, iterkeys


class WannGenomeConfig(object):
    """Sets up and holds configuration information for the DefaultGenome class."""
    allowed_connectivity = [ 'unconnected', 'fs_neat_nohidden', 'fs_neat', 'fs_neat_hidden',
                             'full_nodirect', 'full', 'full_direct',
                             'partial_nodirect', 'partial', 'partial_direct' ]

    def __init__(self, params):
        # Create full set of available activation functions.
        self.activation_defs = ActivationFunctionSet()
        # ditto for aggregation functions - name difference for backward compatibility
        self.aggregation_function_defs = AggregationFunctionSet()
        self.aggregation_defs = self.aggregation_function_defs

        self._params = [ ConfigParameter('num_inputs', int),
                         ConfigParameter('num_outputs', int),
                         ConfigParameter('num_hidden', int),
                         ConfigParameter('feed_forward', bool),
                         ConfigParameter('compatibility_disjoint_coefficient', float),
                         ConfigParameter('compatibility_weight_coefficient', float),
                         ConfigParameter('conn_add_prob', float),
                         ConfigParameter('conn_delete_prob', float),
                         ConfigParameter('node_add_prob', float),
                         ConfigParameter('node_delete_prob', float),
                         ConfigParameter('single_structural_mutation', bool, 'false'),
                         ConfigParameter('structural_mutation_surer', str, 'default'),
                         ConfigParameter('initial_connection', str, 'unconnected'),
                         ConfigParameter('add_conn_roulette_prob', float),
                         ConfigParameter('add_node_roulette_prob', float),
                         ConfigParameter('mutate_conn_roulette_prob', float),
                         ConfigParameter('mutate_act_roulette_prob', float),
                         ConfigParameter('delete_conn_roulette_prob', float),
                         ConfigParameter('delete_node_roulette_prob', float)]

        # Gather configuration data from the gene classes.
        self.node_gene_type = params[ 'node_gene_type' ]
        self._params += self.node_gene_type.get_config_params()
        self.connection_gene_type = params[ 'connection_gene_type' ]
        self._params += self.connection_gene_type.get_config_params()

        # Use the configuration data to interpret the supplied parameters.
        for p in self._params:
            setattr(self, p.name, p.interpret(params))

        # By convention, input pins have negative keys, and the output
        # pins have keys 0,1,...
        self.input_keys = [ -i - 1 for i in range(self.num_inputs) ]
        self.output_keys = [ i for i in range(self.num_outputs) ]

        self.connection_fraction = None

        # Verify that initial connection type is valid.
        # pylint: disable=access-member-before-definition
        if 'partial' in self.initial_connection:
            c, p = self.initial_connection.split()
            self.initial_connection = c
            self.connection_fraction = float(p)
            if not (0 <= self.connection_fraction <= 1):
                raise RuntimeError(
                    "'partial' connection value must be between 0.0 and 1.0, inclusive.")

        assert self.initial_connection in self.allowed_connectivity

        # Verify structural_mutation_surer is valid.
        # pylint: disable=access-member-before-definition
        if self.structural_mutation_surer.lower() in [ '1', 'yes', 'true', 'on' ]:
            self.structural_mutation_surer = 'true'
        elif self.structural_mutation_surer.lower() in [ '0', 'no', 'false', 'off' ]:
            self.structural_mutation_surer = 'false'
        elif self.structural_mutation_surer.lower() == 'default':
            self.structural_mutation_surer = 'default'
        else:
            error_string = "Invalid structural_mutation_surer {!r}".format(
                self.structural_mutation_surer)
            raise RuntimeError(error_string)

        self.node_indexer = None

    def add_activation(self, name, func):
        self.activation_defs.add(name, func)

    def add_aggregation(self, name, func):
        self.aggregation_function_defs.add(name, func)

    def save(self, f):
        if 'partial' in self.initial_connection:
            if not (0 <= self.connection_fraction <= 1):
                raise RuntimeError(
                    "'partial' connection value must be between 0.0 and 1.0, inclusive.")
            f.write('initial_connection      = {0} {1}\n'.format(self.initial_connection,
                                                                 self.connection_fraction))
        else:
            f.write('initial_connection      = {0}\n'.format(self.initial_connection))

        assert self.initial_connection in self.allowed_connectivity

        write_pretty_params(f, self, [ p for p in self._params
                                       if not 'initial_connection' in p.name ])

    def get_new_node_key(self, node_dict):
        if self.node_indexer is None:
            self.node_indexer = count(max(list(iterkeys(node_dict))) + 1)

        new_id = next(self.node_indexer)

        assert new_id not in node_dict

        return new_id

    def check_structural_mutation_surer(self):
        if self.structural_mutation_surer == 'true':
            return True
        elif self.structural_mutation_surer == 'false':
            return False
        elif self.structural_mutation_surer == 'default':
            return self.single_structural_mutation
        else:
            error_string = "Invalid structural_mutation_surer {!r}".format(
                self.structural_mutation_surer)
            raise RuntimeError(error_string)


class WannGenome(neat.DefaultGenome):
    def __init__(self, key):
        super().__init__(key)
        self.fitMax = 0.0

    @classmethod
    def parse_config(cls, param_dict):
        param_dict[ 'node_gene_type' ] = DefaultNodeGene
        param_dict[ 'connection_gene_type' ] = DefaultConnectionGene
        return WannGenomeConfig(param_dict)

    def configure_crossover(self, genome1, genome2, config):
        """ Configure a new genome by crossover from two parent genomes. """
        assert isinstance(genome1.fitness, (int, float))
        assert isinstance(genome2.fitness, (int, float))
        if genome1.rank < genome2.rank:
            parent1, parent2 = genome1, genome2
        else:
            parent1, parent2 = genome2, genome1

        # Inherit connection genes
        for key, cg1 in iteritems(parent1.connections):
            cg2 = parent2.connections.get(key)
            if cg2 is None:
                # Excess or disjoint gene: copy from the fittest parent.
                self.connections[ key ] = cg1.copy()
            else:
                # Homologous gene: combine genes from both parents.
                self.connections[ key ] = cg1.crossover(cg2)

        # Inherit node genes
        parent1_set = parent1.nodes
        parent2_set = parent2.nodes

        for key, ng1 in iteritems(parent1_set):
            ng2 = parent2_set.get(key)
            assert key not in self.nodes
            if ng2 is None:
                # Extra gene: copy from the fittest parent
                self.nodes[ key ] = ng1.copy()
            else:
                # Homologous gene: combine genes from both parents.
                self.nodes[ key ] = ng1.crossover(ng2)

    def mutate(self, config):
        """ Mutates this genome. """
        topology_roulette = np.array(
            (config.add_conn_roulette_prob, config.add_node_roulette_prob,
             config.mutate_conn_roulette_prob, config.mutate_act_roulette_prob, config.delete_conn_roulette_prob,
             config.delete_node_roulette_prob))

        spin = np.random.rand() * np.sum(topology_roulette)
        slot = topology_roulette[ 0 ]
        choice = topology_roulette.size
        for i in range(1, topology_roulette.size):
            if spin < slot:
                choice = i
                break
            else:
                slot += topology_roulette[ i ]

        # Add Connection
        if choice is 1:
            self.mutate_add_connection(config)

        # Add Node
        elif choice is 2:
            self.mutate_add_node(config)

        # Mutate Connection
        elif choice is 3:
            if len(self.connections) > 0:
                random.choice(list(self.connections.values())).mutate(config)

        # Mutate Activation
        elif choice is 4:
            if len(self.nodes) > 0:
                random.choice(list(self.nodes.values())).mutate(config)

        # Delete connection
        elif choice is 5:
            self.mutate_delete_connection()

        # Delete node
        elif choice is 6:
            self.mutate_delete_node(config)